# -*- coding:utf-8 -*-
from dataclasses import dataclass

import pytest

from test.model import Student
from ysql.entity import _parse_attr_value, Entity, Constraint as cs


def test_simple_create():
    @Entity  # 表明这是数据（表）类
    @dataclass
    class Student:  # 类名即表名，不区分大小写
        score: float
        name: str = '好家伙', cs.not_null  # 同时设置默认值和约束，以逗号分开即可
        student_id: int = cs.auto_primary_key  # Constraint类提供多种字段约束

    student = Student(name="李华", score=95.6)
    print(student)


def test_without_check_type():
    @Entity
    @dataclass
    class StudentNoCheck:
        name: str
        age: int = 10
        phone: int = cs.not_null
        weight: float = 50.0, cs.not_null
        height: float = cs.not_null, cs.unique
        address: str = 'hit', cs.not_null, cs.unique
        student_id: int = cs.auto_primary_key

    student = StudentNoCheck(name=f'李华', age=1, phone=123456789,
                             weight=50.0, height=100.0, address=f'hit')

    print(student)


def test_type_float_to_int():
    with pytest.raises(TypeError) as error:
        Student(name='李华1', age=1.1, phone=123456789,
                weight=50.0, height=100.0, address=f'hit',
                student_id=1)
    assert "实例化" in str(error.value)
    assert "类时," in str(error.value)


def test_check_type():
    students = [
        Student(name=f'李华{i}', age=i, phone=123456789,
                weight=50.0, height=100.0 + i, address=f'hit{i}')
        for i in range(1, 100)
    ]
    for student in students:
        print(student)


def test_no_default_value():
    student = Student(name=f'李华1', phone=123456789,
                      height=100.0, student_id=1)
    assert student.age == _parse_attr_value(Student.age)
    assert student.weight == _parse_attr_value(Student.weight)
    assert student.address == _parse_attr_value(Student.address)


def test_change_position():
    @Entity(check_type=True)
    @dataclass
    class StudentChangePosition:
        name: str
        age: int = 10
        phone: int = cs.not_null
        weight: float = cs.not_null, 50.0
        height: float = cs.not_null, cs.unique
        address: str = cs.not_null, 'hit', cs.unique
        student_id: int = cs.auto_primary_key

    student = StudentChangePosition(name=f'李华1', phone=123456789,
                                    height=100.0, student_id=1)
    assert student.age == _parse_attr_value(Student.age)
    assert student.weight == _parse_attr_value(Student.weight)
    assert student.address == _parse_attr_value(Student.address)


def test_multiple_default_value():
    @Entity(check_type=True)
    @dataclass
    class StudentMultipleDefaultValue:
        name: str
        age: int = 10
        phone: int = cs.not_null
        weight: float = cs.not_null, 50.0
        height: float = cs.not_null, cs.unique
        address: str = cs.not_null, 'hit', cs.unique, 'hit'
        student_id: int = cs.auto_primary_key

    with pytest.raises(TypeError) as error:
        student = StudentMultipleDefaultValue(name=f'李华1', phone=123456789,
                                              height=100.0, student_id=1)
    assert "数据类的属性值不满足要求，只允许1个基本类型默认值和Constraint类约束" in str(error.value)


def test_repeat_constraint():
    @Entity(check_type=True)
    @dataclass
    class StudentRepeatConstraint:
        name: str
        age: int = 10
        phone: int = cs.not_null
        weight: float = cs.not_null, 50.0
        height: float = cs.not_null, cs.unique, cs.unique
        address: str = cs.not_null, 'hit', cs.unique
        student_id: int = cs.auto_primary_key

    student = StudentRepeatConstraint(name=f'李华1', phone=123456789,
                                      height=100.0, student_id=1)


def test_None():
    @Entity(check_type=True)
    @dataclass
    class StudentNone:
        name: str
        age: int = None
        phone: int = cs.not_null, None
        weight: float = None, cs.not_null
        height: float = cs.not_null, cs.unique
        address: str = cs.not_null, None, cs.unique
        student_id: int = cs.auto_primary_key

    student = StudentNone(name=f'李华1', height=100.0, student_id=1)
    assert student.age is None
    assert student.phone is None
    assert student.address is None
    assert student.height == 100.0


def test_type_bool():
    @Entity(check_type=True)
    @dataclass
    class Student:
        name: str
        if_18: bool
        contact_info: float
        student_id: int = cs.auto_primary_key, cs.comment('学生id')

    Student = Student(name='李华', if_18=True, contact_info=123.0)


if __name__ == '__main__':
    pytest.main(["-v", __file__, "--log-cli-level=INFO"])



# -*- coding: utf-8 -*-
import copy
from collections import namedtuple
from dataclasses import make_dataclass, dataclass

from ysql.database import log
from ysql.entity import Constraint, _parse_constraints, _ENABLED_TYPES, is_entity, _ENABLED_TYPES_WITHOUT_NONE

# ====================================================================================================================
# 模块常量
TABLE_SUBSTITUTE = '__'  # 约定的表名代替符，在sql语句中凡是涉及本表表名的，均可采取此代替符，内部可自动替换。
RETURN_FLAG = ('limit 1', 'limit 1;')  # 在sql语句中包含此字符，则直接返回单条记录，不再返回列表。
_DATACLASS = 'dataclass'
_RECORD = 'Record'


# ====================================================================================================================
# 装饰器方法
def Dao(entity, return_type=tuple, auto_single_record: bool = True):
    """对数据访问类使用的装饰器，用于绑定数据类和数据访问类。

    Args:
        entity: 该数据访问类对应的数据类。
        return_type: [tuple, dict, dataclass, your_entity], 设置select语句查询结果的数据类型。
        auto_single_record: 自动识别并返回单条select记录。

    Example:

        @Entity
        @dataclass
        class Student:  # 定义一个数据类
            name: str
            score: float

        @Dao(Student)  # 通过Dao装饰器绑定对应的数据类
        class DaoStudent:  # 定义一个数据访问类
            ...

    """

    def decorator(cls):
        # 新增属性
        cls.cursor = None
        cls.entity = entity
        cls.auto_single_record = auto_single_record
        cls.return_type = return_type if return_type != dataclass else _DATACLASS
        # 新增方法
        cls.execute = execute
        cls.update_cursor = update_cursor
        cls._create_table = _create_table
        cls._generate_sql_create_table = _generate_sql_create_table
        return cls

    return decorator


def Sql(sql: str, return_type=None):
    """执行sql语句的装饰器，传入sql语句的同时会自动实现被装饰函数。

    Args:
        sql: 固定字符串sql语句。
        return_type: [tuple, dict, dataclass, your_entity], 设置select语句查询结果的数据类型，
                    此参数优先级更高，若Dao和Sql同时设置该参数，则遵循Sql的return_type。

    Example:

        # 对于固定的sql语句可以直接传入Sql装饰器。
        # 此段代码无法直接运行，需要首先连接数据库并插入记录才可运行。

        @Entity
        @dataclass
        class Student:
            name: str
            score: float
            student_id: int = Constraint.auto_primary_key

        @Dao(Student)
        class DaoStudent:

            @Sql("select * from student where student_id=?;")
            def get_student(self, student_id):
                pass  # 此处并非以pass示意省略，而是Sql装饰器会自动实现该函数，因此实际使用时均只需要pass即可。

        dao = DaoStudent()
        result = dao.get_student(student_id=1)

        # 将以列表形式返回查询结果，其中每条记录的数据格式默认为dataclass格式。
        # result1: [Record(name, score, student_id), ...]
        # 查询部分字段时，仅以查询的字段生成相应的dataclass。

    !Note:
        1.对于固定的静态sql语句，通过Sql装饰器传递sql语句，被装饰器函数传递数据。
        2.对于复杂的动态sql语句，不应使用Sql装饰器，应使用dao.execute方法。

    """

    def decorator(func):  # noqa

        def wrapper(self, *args, **kwargs):
            if not type(sql) == str:
                raise ValueError(
                    f"传入的sql参数应该是 'str' 类型，但得到的是 '{type(sql).__name__}'类型")

            final_sql = __check_table_substitute(sql=sql, entity=self.entity)
            log.debug(f"生成的执行sql:{final_sql}")

            self.cursor.execute(final_sql, (*args, *kwargs.values()))
            return __fetch_result(self=self,
                                  sql=final_sql,
                                  cursor=self.cursor,
                                  return_type=return_type)

        return wrapper

    if callable(sql):
        raise ValueError(
            f"Sql装饰器需要传入一个 'str' 类型的参数")

    return decorator


def Insert(func):  # noqa
    """执行插入功能的装饰器，会自动生成插入sql语句，以及自动实现被装饰函数。

    Returns:
        自动返回刚插入记录的自增主键（如果使用自增主键）。

    Example:

        # 此段代码无法直接运行，需要首先连接数据库才可运行。

        @Entity
        @dataclass
        class Student:
            name: str
            score: float
            student_id: int = Constraint.auto_primary_key

        @Dao(Student)
        class DaoStudent:

            @Insert  # 无需传递任何参数或者sql语句。
            def insert(self, entity):
                pass  # 此处并非以pass示意省略，而是Insert装饰器会自动实现该函数，因此实际使用时均只需要pass即可。

        dao = DaoStudent()
        # 插入单条记录
        student = Student(name='Bob', score=95.5)  # 将整条记录以数据类的形式插入数据库，避免了同时使用多个参数的麻烦。
        dao.insert(entity=student)
        # 批量插入多条记录
        students = [Student(name='Bob', score=i) for i in range(100)]
        dao.insert(entity=students)
    """

    def wrapper(self, entity):
        """entity参数是传入的数据类实例（对象），而self.entity是定义的数据类（类）"""
        # 获取entity类的属性名和类型
        fields = [field_name for field_name, _ in self.entity.__annotations__.items()]

        # 定义需过滤的属性
        ignore_fields = {Constraint.auto_primary_key.constraint[0],
                         Constraint.ignore.constraint[0]}
        ignore_fields = {field_name for field_name in fields
                         if ignore_fields & set(_parse_constraints(attr_value=getattr(self.entity, field_name, None)))}
        # 过滤属性
        filtered_fields = [field_name for field_name in fields
                           if field_name not in ignore_fields]

        sql = f"insert into {self.entity.__name__} " \
              f"({', '.join(field_name for field_name in filtered_fields)}) " \
              f"values ({', '.join('?' for _ in filtered_fields)});"
        log.debug(f"生成的插入sql:{sql}")

        if type(entity) in (list, tuple):
            values = ([getattr(item, field_name) for field_name in filtered_fields]
                      for item in entity)
            self.cursor.executemany(sql, values)
            return

        cls_fields = set(field_name for field_name, _ in entity.__annotations__.items()
                         if field_name not in ignore_fields)
        if cls_fields == set(filtered_fields):
            values = [getattr(entity, field_name) for field_name in filtered_fields]
            log.debug(f"获取到的插入参数:{values}")
            self.cursor.execute(sql, values)
            return self.cursor.lastrowid

        else:
            raise TypeError(
                f"Insert的被装饰函数的参数类型应该是:\n"
                f"1.绑定的Entity数据类的实例\n"
                f"2.包含多个数据类的序列，'list' or 'tuple' "
                f"但得到的是 {type(entity).__name__}")

    return wrapper


# ====================================================================================================================
# 提前实现了一些可用方法的功能类，建议继承。
class DaoFunc:
    """
    Example:
        @Entity
        @dataclass
        class Student:
            name: str
            score: float
            student_id: int = Constraint.auto_primary_key

        @Dao(Student)
        class DaoStudent(DaoFunc):
            pass

    """

    @Insert
    def insert(self, entity):
        """插入单条或多条数据记录

        Args:
            entity: 数据类的实例，或包含多个数据类的序列。

        Returns:
            自动返回刚插入记录的自增主键（如果使用自增主键）。

        """
        pass

    def execute(self, sql: str, args=None, return_type=None):
        """执行sql语句"""
        pass


# ====================================================================================================================
# Dao类新增的方法
def execute(self, sql: str, args=None, return_type=None):
    """执行sql语句"""
    sql = __check_table_substitute(sql=sql, entity=self.entity)
    if args is None:
        self.cursor.execute(sql)
    elif type(args) in _ENABLED_TYPES:
        self.cursor.execute(sql, (args,))
    elif type(args) in (list, tuple):
        self.cursor.execute(sql, __flatten_args(args))

    return __fetch_result(self=self, sql=sql, cursor=self.cursor, return_type=return_type)


def _create_table(self):
    """依据entity创建表"""
    sql = self._generate_sql_create_table()
    self.cursor.execute(sql)


def update_cursor(self, cursor):
    """更新dao中的游标"""
    self.cursor = cursor


def _generate_sql_create_table(self):
    """自动生成建表语句"""
    table_name = self.entity.__name__.lower()
    # 获取字段名称和类型的字典
    fields = self.entity.__annotations__

    field_definitions = []
    foreign_key_constraints = []

    for field_name, field_type in fields.items():
        # 获取字段的SQL类型
        sql_type = __convert_to_sql_type(field_type)
        # 获取字段的约束条件
        constraints = _parse_constraints(attr_value=getattr(self.entity, field_name, None))
        # 跳过忽略属性
        if Constraint.ignore.constraint[0] in constraints:
            continue

        foreign_key_constraint = [item for item in constraints
                                  if isinstance(item, tuple)]
        constraints = [item for item in constraints
                       if item not in foreign_key_constraint]
        # 有外键
        if len(foreign_key_constraint) == 1:
            foreign_key_constraint = __generate_sql_foreign_key(field_name=field_name,
                                                                constraint=foreign_key_constraint[0])
            foreign_key_constraints.append(foreign_key_constraint)

        elif len(foreign_key_constraint) > 1:
            raise TypeError(
                '一个属性只能指定一个外键')

        # 合并其他约束条件
        constraint = " ".join(constraints)

        # 拼接字段定义
        field_definition = f"{field_name} {sql_type} {constraint}"
        field_definitions.append(field_definition)
    # 将外键约束列表添加到字段定义列表的末尾
    field_definitions.extend(foreign_key_constraints)
    # 默认为严格的数据类型约束
    sql_create_table = f"CREATE TABLE IF NOT EXISTS {table_name} ({', '.join(field_definitions)});"
    log.debug(f"生成的建表SQL语句：{sql_create_table}")
    return sql_create_table


# ====================================================================================================================
# 模块内的静态方法
def __convert_to_sql_type(python_type):
    """转换python注释类型为sql类型"""
    if python_type == int:
        return "INTEGER"
    elif python_type == str:
        return "TEXT"
    elif python_type == float:
        return "REAL"
    elif python_type == bool:
        return "BOOL"
    elif python_type == bytes:
        return "BLOB"

    raise ValueError(
        f"ysql不支持该python数据类型: {python_type}")  # noqa


def __generate_sql_foreign_key(constraint: tuple, field_name: str):
    """生成外键约束的sql语句"""
    log.debug(f'获得的外键约束：{constraint}')
    foreign_key_entity, foreign_key_name, foreign_key_delete_link, foreign_key_update_link = constraint

    if foreign_key_delete_link is None and foreign_key_update_link is None:
        return f"FOREIGN KEY ({field_name}) " \
               f"REFERENCES {foreign_key_entity}({foreign_key_name})"

    # 存在表关联的情况
    elif foreign_key_delete_link is not None and foreign_key_update_link is None:
        return f"FOREIGN KEY ({field_name}) " \
               f"REFERENCES {foreign_key_entity}({foreign_key_name}) ON DELETE {foreign_key_delete_link}"
    elif foreign_key_delete_link is None and foreign_key_update_link is not None:
        return f"FOREIGN KEY ({field_name}) " \
               f"REFERENCES {foreign_key_entity}({foreign_key_name}) ON UPDATE {foreign_key_update_link}"
    elif foreign_key_delete_link is not None and foreign_key_update_link is not None:
        return f"FOREIGN KEY ({field_name}) " \
               f"REFERENCES {foreign_key_entity}({foreign_key_name}) " \
               f"ON DELETE {foreign_key_delete_link} ON UPDATE {foreign_key_update_link} "


def __check_table_substitute(sql: str, entity):
    """检查并替换表名代替符"""
    return sql.replace(TABLE_SUBSTITUTE, entity.__name__.upper())


def __flatten_args(args):
    """一维展开不定参数，并返回元组形式"""

    def flatten_args(arg):
        if type(arg) in (list, tuple):
            # 如果是列表或元组，递归展开每个元素
            return [item for sublist in map(flatten_args, arg) for item in sublist]
        elif type(arg) == dict:
            # 如果是字典，取出所有值并递归展开
            return flatten_args(list(arg.values()))
        elif type(arg) in _ENABLED_TYPES:
            # 否则返回单个值
            return [arg]
        else:
            raise ValueError(
                f"args参数的类型应该是 'list' or 'tuple' or 'dict' ,但得到的是 '{type(arg).__name__}'"
            )

    return tuple(flatten_args(args))


# --------------------------------------------------------------------------------------------------------------------
# 设置返回结果格式
def __return_entity(entity, cursor, result, is_list=True):
    """返回查询结果为传入的entity数据类"""
    cursor_fields = [column[0] for column in cursor.description]
    entity_fields = [attr_name for attr_name, _ in entity.__annotations__.items()]
    # 检查是否所有cursor字段都在entity中
    if not all(field in entity_fields for field in cursor_fields):
        raise ValueError("Cursor fields are not a subset of entity fields.")

    def create_entity_fast(row):
        _entity = entity.__new__(entity)
        entity.__entity_init__(_entity, *row)
        return _entity

    def create_entity(row):
        _entity = entity.__new__(entity)
        for field, value in zip(field_mapping, row):
            setattr(_entity, field, value)
        return _entity

    if cursor_fields == entity_fields:
        create = create_entity_fast
    else:
        # 映射cursor字段到entity字段
        field_mapping = [field for field in cursor_fields if field in entity_fields]
        create = create_entity

    if is_list:
        return [create(row) for row in result]
    return create(result)


def __return_dataclass(cursor, result, is_list=True):
    """返回询结果为dataclass"""
    fields = [column[0] for column in cursor.description]
    Record = make_dataclass(_RECORD, fields)
    if is_list:
        return [Record(*row) for row in result]
    return Record(*result)


def __return_dict(cursor, result, is_list=True):
    """返回查询结果为字典"""
    fields = [column[0] for column in cursor.description]
    if is_list:
        return [dict(zip(fields, item)) for item in result]
    return dict(zip(fields, result))


def __return_namedtuple(cursor, result, is_list=True):
    """返回查询结果为具名元组"""
    fields = [column[0] for column in cursor.description]
    Record = namedtuple(_RECORD, fields)
    if is_list:
        return [Record._make(row) for row in result]  # noqa
    return Record._make(result)  # noqa


def __return_single_field(result, return_type, is_list=True):
    """提取出单字段并返回，并处理bool类型"""
    if return_type != bool:
        if is_list:
            return [item[0] if item is not None else item
                    for item in result]
        return result[0] if result is not None else result
    else:
        if is_list:
            return [bool(item[0]) if item is not None else item
                    for item in result]
        return bool(result[0]) if result is not None else result


def __is_auto_single_record(sql: str) -> bool:
    """判断sql语句中是否明确仅取一条记录"""
    return any([sql.endswith(flag) for flag in RETURN_FLAG])


def __fetch_result(self, sql, cursor, return_type):
    """确定返回类型"""
    lower_sql = sql.lower()
    # 确定返回类型
    _return_type = copy.deepcopy(self.return_type)
    if return_type is not None:
        _return_type = return_type

    if self.auto_single_record and __is_auto_single_record(lower_sql):
        is_list = False
        result = cursor.fetchone()
    else:
        is_list = True
        result = cursor.fetchall()

    # 不做转换
    if not result or _return_type == tuple:
        return result

    elif is_entity(_return_type):
        return __return_entity(entity=_return_type, cursor=cursor, result=result, is_list=is_list)

    elif _return_type in _ENABLED_TYPES_WITHOUT_NONE:
        return __return_single_field(result=result, return_type=_return_type, is_list=is_list)

    elif _return_type == _DATACLASS:
        return __return_dataclass(cursor=cursor, result=result, is_list=is_list)

    elif _return_type == dict:
        return __return_dict(cursor=cursor, result=result, is_list=is_list)

    else:
        raise TypeError(
            f"Dao的查询结果格式设置错误，应该属于: 'tuple' 'dict' 'dataclass' 'entity'，但得到的是: {_return_type}")



# -*- coding: utf-8 -*-
import logging
import os
import sqlite3

# ====================================================================================================================
# 日志配置
log = logging.getLogger(__name__)


def setup_logging(debug=False):
    log.setLevel(logging.DEBUG if debug else logging.CRITICAL)

    # 控制台处理器
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
    log.addHandler(console_handler)


setup_logging(debug=False)


# ====================================================================================================================
class MetaDatabase:
    """元数据库类，提供库级的操作

    由于动态元编程无法使用代码提示，因此采取继承写法。

    Example:

        @Entity
        @dataclass
        class Student:  # 定义一个数据类
            name: str
            student_id: int = Constraint.auto_primary_key

        @Dao(Student)
        class DaoStudent:  # 定义一个数据访问类

            @Sql("select * from student where student_id=?;")
            def get_student(self, student_id):
                pass

        class Database(MetaDatabase):  # 定义一个数据库类，继承元数据库类
            dao1 = DaoStudent()  # 将各个数据访问类实例化为类中静态变量，集中管理，统一对外。
            dao2 = ...
            dao3 = ...

        db = Database(db_path='test0.db')  # 实例化数据库类，并传入数据库路径
        db.connect()  # 连接数据库
        db.create_tables()  # 创建数据表
        db.commit()  # 提交更改

    """

    # ================================================================================================================
    # 提供的可调用方法
    def connect(self, use_multithreading=False):
        """连接数据库"""
        self.__check_path(self.db_path)
        self.connection = sqlite3.connect(self.db_path, check_same_thread=not use_multithreading)
        self.cursor = self.connection.cursor()
        self.__update_cursor()
        log.debug('连接成功')

    def disconnect(self):
        """断开数据库连接"""
        if self.connection:
            self.connection.close()
            log.debug('已断开数据库连接')

    def create_tables(self):
        """当表不存在时，才会创建数据表。因此可以反复调用该方法，而不会产生错误。"""
        for dao in self.dao_list:
            dao._create_table()  # noqa
        self.commit()
        log.debug('已创建全部表')

    def commit(self):
        """提交事务"""
        self.connection.commit()
        log.debug('提交事务成功')

    def rollback(self):
        """回滚事务"""
        self.connection.rollback()
        log.debug('回滚事务成功')

    def execute(self, statement: str):
        """给外部提供的直接执行sql的接口，避免了再调用内部的connection"""
        self.connection.execute(statement)
        log.debug(f"执行了sql语句：{statement}")

    # ================================================================================================================
    # 内部方法
    def __new__(cls, *args, **kwargs):
        # 获取子类的所有属性
        subclass_attrs = dir(cls)
        # 初步筛选静态变量（不包括方法和特殊属性）
        static_attrs = [attr for attr in subclass_attrs
                        if not callable(getattr(cls, attr)) and not attr.startswith("__")]
        # 根据是否具有entity属性筛选出最终dao属性
        dao_list = [getattr(cls, attr) for attr in static_attrs
                    if hasattr(getattr(cls, attr), "entity")
                    and hasattr(getattr(cls, attr), "update_cursor")]
        # 赋值dao_list为类属性，以便在类内部可见
        cls.dao_list = dao_list

        return super().__new__(cls)

    def __init__(self, db_path: str):
        self.connection = None
        self.cursor = None
        self.db_path = db_path

    # ================================================================================================================
    # 类内使用的方法
    def __update_cursor(self):
        for dao in self.dao_list:
            dao.update_cursor(cursor=self.cursor)

    @staticmethod
    def __check_path(path):
        """检查并确保目录存在"""
        db_folder = os.path.dirname(path)
        if db_folder != '' and not os.path.exists(db_folder):
            os.makedirs(db_folder, exist_ok=True)
            log.debug(f'数据库路径父目录不存在，已自动创建')



# -*- coding: utf-8 -*-
from typing import Union, NamedTuple


# ====================================================================================================================
# 模块常量
class _Constraint(NamedTuple):
    """内部维护的约束类，存储实际约束"""
    constraint: tuple


_PRIMARY_KEY = _Constraint(constraint=("PRIMARY KEY",))
_AUTO_PRIMARY_KEY = _Constraint(constraint=("PRIMARY KEY AUTOINCREMENT",))
_NOT_NULL = _Constraint(constraint=("NOT NULL",))
_UNIQUE = _Constraint(constraint=("UNIQUE",))
_IGNORE = _Constraint(constraint=("ignore",))

_NO_ACTION = 'NO ACTION'
_RESTRICT = 'RESTRICT'
_CASCADE = 'CASCADE'
_SET_NULL = 'SET NULL'
_SET_DEFAULT = 'SET DEFAULT'

_ENABLED_TYPES = (int, str, float, bytes, bool, type(None))
_ENABLED_TYPES_WITHOUT_NONE = (int, str, float, bytes, bool)
_ENABLED_UNION = Union[_ENABLED_TYPES]
_ENTITY_INIT = "__entity_init__"


# ====================================================================================================================
def Entity(check_type: bool = False):
    """数据类dataclass的装饰器。

    在定义dataclass时，同时将约束条件和默认值赋予类属性。
    通过改造原生dataclass的init方法和getattribute方法，实现sql建表时约束条件的解析，以及正常使用被装饰dataclass时属性值的获取。

    Args:
        check_type(bool): 传入True -> 实例化dataclass时检查属性值和类型注释是否一致，不一致将触发类型错误。否则不检查。

    Example1:

        @Entity
        @dataclass
        class Student:  # 类名即表名，不区分大小写
            name: str  # 字段名以及字段类型
            score: float = 100

    Example2:

        @Entity(check_type=True)  # 启用类型检查
        @dataclass
        class Student:
            name: str  # 字段名以及字段类型
            score: float = 100.0
            # Example1中的score属性虽然注释为float，而实际默认值为int，但仍然可以正常工作（由于python和sqlite都是宽松的类型约束）。
            # 若使用check_type，将进行严格的检查，类型不一致的情况（如Example1）会抛出异常。

    !Note:
        在实用角度上@dataclass应该合并到@Entity内部，这样定义数据类时只需要使用一个装饰器，并且不需要初级使用者关心什么是dataclass，
        但经过测试发现，如果不显式的使用@dataclass装饰器，在实例化数据类时Pycharm将无法给出代码提示，这是不可忍受的。
        并且经过测试发现只有Pycharm2020及之前的版本可以正确给出代码提示，高于2020版存在代码提示的bug，详见:
        https://intellij-support.jetbrains.com/hc/en-us/community/posts/4421575751442-Code-Completion-doesn-t-work-for-class-functions-decorated-with-decorators-that-return-inner-functions

    """

    def decorator(cls):
        orig_get_attr = cls.__getattribute__
        orig_init = cls.__init__

        def get_attr(self, name):
            """重写被装饰dataclass的取值方法"""
            # 只有访问定义的属性时才进一步处理
            if name in set(attr_name for attr_name, attr_type in cls.__annotations__.items()):
                return _parse_attr_value(orig_get_attr(self, name))
            # 访问内部属性则返回原始值
            return orig_get_attr(self, name)

        def init_and_check_type(self, *args, **kwargs):
            """重写被装饰dataclass的初始化方法"""
            orig_init(self, *args, **kwargs)
            for attr_name, attr_type in cls.__annotations__.items():
                # 再检查属性值的类型与类型注解是否一致
                attr_value = get_attr(self, attr_name)
                # 不检查默认值为None的属性
                if attr_value is None:
                    continue
                elif type(attr_value) != attr_type:
                    raise TypeError(
                        f"实例化'{cls.__name__}'类时,"
                        f"'{attr_name}'属性的类型应该是 '{attr_type.__name__}' ,"
                        f"但得到的是 '{type(attr_value).__name__}'")

        # 由于无参装饰器的特性，在无参使用该装饰器时，check_type的值被覆盖为cls，因此必须显式的与True进行判断
        if check_type == True:  # noqa
            cls.__init__ = init_and_check_type
        setattr(cls, _ENTITY_INIT, orig_init)
        cls.__getattribute__ = get_attr
        return cls

    # 无参使用该装饰器时
    if callable(check_type):
        return decorator(cls=check_type)  # noqa

    return decorator


def is_entity(obj):
    return hasattr(obj, _ENTITY_INIT)


class Constraint:
    """对外开放的各种字段约束。

    Example:

        @Entity
        @dataclass
        class Student:
            name: str  # 可以不使用约束
            score: float = 100  # 可以只赋予默认值
            address: str = 'HIT', Constraint.not_null  # 同时使用默认值和约束，需要以逗号分隔开
            student_id: int = Constraint.auto_primary_key, Constraint.not_null  # 同时使用多条约束，需要以逗号分隔开

    !Note:
        建议导入Constraint类时使用别名，可以有效简化存在大量约束的使用场景。

        Example:

            from ysql import Constraint as cs

            @Entity
            @dataclass
            class Student:
                name: str
                score: float = 100
                address: str = 'HIT', cs.not_null
                student_id: int = cs.auto_primary_key, cs.not_null

    """

    # =================================================================================================================
    # 1.可直接使用的约束常量
    primary_key = _PRIMARY_KEY  # 主键
    auto_primary_key = _AUTO_PRIMARY_KEY  # 自增主键
    not_null = _NOT_NULL  # 非空
    unique = _UNIQUE  # 唯一

    # 非sql的特殊约束
    ignore = _IGNORE  # 用于建表时忽略某字段/属性

    # 针对外键的约束
    no_action = _NO_ACTION
    cascade = _CASCADE
    set_null = _SET_NULL
    restrict = _RESTRICT
    set_default = _SET_DEFAULT

    # =================================================================================================================
    # 2.需要外部传值的约束
    @staticmethod
    def default(default_value: _ENABLED_UNION):
        """默认值约束

        Args:
            default_value: 该字段在sql中的默认值，与定义数据类时使用默认值作用类似。

        Example:

            @Entity
            @dataclass
            class Student:
                name: str
                score1: float = 100
                score2: float = Constraint.default(100)  # score1和score2的作用类似
                student_id: int = Constraint.auto_primary_key

        """
        if type(default_value) in _ENABLED_TYPES:
            return _Constraint(constraint=(f'DEFAULT {default_value}',))  # noqa
        raise TypeError(
            f"dataclass属性默认值允许的数据类型: 'int', 'str', 'float', 'bytes',"
            f"但得到的是 '{type(default_value).__name__}'"
        )

    @staticmethod
    def check(check_condition: str):
        """条件约束

        Args:
            check_condition: 具体条件

        Example:

            @Entity
            @dataclass
            class Student:
                name: str
                score: float = Constraint.check('score > 60')  # 需要填写字段的字符形式名称
                student_id: int = Constraint.auto_primary_key

        """
        if type(check_condition) == str:
            return _Constraint(constraint=(f'CHECK({check_condition})',))  # noqa
        raise TypeError(
            f"对dataclass属性使用条件约束时，允许的数据类型: 'str',"
            f"但得到的是 '{type(check_condition).__name__}'")

    @staticmethod
    def foreign_key(entity, field, delete_link=None, update_link=None):
        """外键约束

        Args:
            entity: 外键所在的数据类（父表）
            field: 外键对应的数据类属性
            delete_link: 级联删除方式
            update_link: 级联更新方式

        Example:

            @Entity
            @dataclass
            class Student:  # 父表
                name: str = Constraint.not_null
                student_id: int = Constraint.auto_primary_key

            @Entity
            @dataclass
            class Score:  # 子表
                score: float
                score_id: int = Constraint.auto_primary_key
                # 对student_id字段设置外键关联
                student_id: int = Constraint.foreign_key(entity=Student,
                                                         field='student_id',
                                                         delete_link=Constraint.cascade,
                                                         update_link=Constraint.cascade)

        """
        return _Constraint(constraint=((entity.__name__.lower(), field, delete_link, update_link),))  # noqa

    @staticmethod
    def comment(comment: str):
        """字段注释

        Args:
            comment: 具体注释。注意，在sqlite中只能通过DDL(Data Definition Language)查看。

        Example:

            @Entity
                @dataclass
                class Student:
                    name: str = Constraint.comment('学生姓名')
                    student_id: int = Constraint.auto_primary_key, Constraint.comment('学生id')

        """
        if type(comment) == str:
            # 目前仅支持sqlite的注释格式
            return _Constraint(constraint=(f'-- {comment}\n',))  # noqa
        raise TypeError(
            f"对dataclass属性使用sql注释时，允许的数据类型: 'str',"
            f"但得到的是 '{type(comment).__name__}'")


# ====================================================================================================================
# 模块方法
def _parse_constraints(attr_value) -> list[str, tuple]:
    """从属性原始值中解析出约束条件。
    Args:
        attr_value: 属性原始值(同时包含默认值和约束条件，但仅在定义数据类时有效，实例化后不再包含约束条件)，如Student.name

    Returns:
        list[约束1, 约束2, 约束3, ...]

    Example:

        @Entity
        @dataclass
        class Student:
            name: str = Constraint.comment('学生姓名')
            student_id: int = Constraint.auto_primary_key, Constraint.comment('学生id')

        print(parse_constraints(Student.name))  # output: ['-- 学生姓名\n']
        print(parse_constraints(Student.student_id))  # output: ['PRIMARY KEY AUTOINCREMENT', '-- 学生id\n']

    """
    if not isinstance(attr_value, tuple):
        return []

    if len(attr_value) == 1 and isinstance(attr_value, _Constraint):
        return list(attr_value.constraint)

    elif len(attr_value) > 1:
        constraints = []

        for item in attr_value:
            if not isinstance(item, _Constraint):
                continue

            if item in constraints:
                raise TypeError(
                    f"重复使用了约束条件: {item.constraint}")

            constraints.append(item.constraint[0])

        return constraints

    else:
        raise TypeError(
            f"属性值的数据结构不满足要求，无法解析出sql约束\n传入的值: {attr_value}")


def _parse_attr_value(attr_value) -> _ENABLED_UNION:
    """从属性原始值中过滤约束条件，解析出真正的属性值。

    Args:
        attr_value: 属性原始值(同时包含默认值和约束条件，但仅在定义数据类时有效，实例化后不再包含约束条件)，如Student.name

    Returns:
        真正的属性值

    Example:

        @Entity
        @dataclass
        class Student:
            name: str
            student_id: int = Constraint.auto_primary_key

        bob = Student(name='Bob')
        print(parse_attr_value(bob.name))  # output: 'Bob'
        print(parse_attr_value(bob.student_id))  # output: None

    !Note:
        如果对数据类的某个属性使用了约束，且未指定默认值，而在实例化时不传入该属性的值。这不会引发实例化错误，该属性会存在一个默认的None值。
        因此在实例化时请仔细检查是否对每个需要赋值的属性都完成了赋值，避免遗漏某个属性的赋值。

        Example:

            @Entity
            @dataclass
            class Student:
                name: str
                student_id: int = Constraint.auto_primary_key  # 不指定默认值

            bob = Student(name='Bob')  # 实例化时不对student_id传值
            print(bob.student_id)  # output: None

    """
    # 单个值
    if type(attr_value) in _ENABLED_TYPES:
        return attr_value
    # 单个约束
    if isinstance(attr_value, _Constraint) and len(attr_value) == 1:
        return None
    # 排除非元组情况
    if not isinstance(attr_value, tuple):
        raise TypeError(
            f"数据类的属性值{attr_value}不满足要求，只允许1个基本类型默认值和Constraint类约束")

    constraints = [value for value in attr_value if isinstance(value, _Constraint)]
    values = [value for value in attr_value if type(value) in _ENABLED_TYPES]

    # 存在基本类型和约束之外的值
    if len(constraints) + len(values) != len(attr_value):
        raise TypeError(
            "数据类的属性值不满足要求，只允许1个基本类型默认值和Constraint类约束")
    # 存在一个真正属性值
    if len(values) == 1:
        return values[0]
    # 全是约束
    elif len(values) == 0:
        return None
    else:
        raise TypeError(
            "数据类的属性值不满足要求，只允许1个基本类型默认值和Constraint类约束")
